#![forbid(unsafe_code)]
#![feature(new_uninit)]

use std::mem::MaybeUninit;

use gitea::PullWh;
use log;
use teloxide_core::{prelude::*, types::{Recipient, Chat}};
use teloxide::Bot;
use tide::{Server, Request};
use tokio::{fs, join, sync::Mutex};

mod gitea;

const POLL_TIME: &str = dotenvy_macro::dotenv!("POLL_TIME", "Poll time is not set!");
const CHAT_ID: &str = dotenvy_macro::dotenv!("CHAT_ID", "Chat ID is not set!");
const TG_KEY: &str = dotenvy_macro::dotenv!("TG_KEY", "Telegram key is not set!");
const GITEA_URL: &str = dotenvy_macro::dotenv!("GITEA_URL", "Gitea url is not set!");

const WH_SECRET: &str = dotenvy_macro::dotenv!("WH_SECRET", "Webhook secret is not set!");
const WH_LISTEN_URL: &str = dotenvy_macro::dotenv!("WH_URL", "Webhook listen URL is not set!");

#[derive(Debug, Clone)]
struct SharedState {
    bot: Bot,
    chat: Chat
}

async fn start_tg() -> (Bot, Chat) {
    log::info!("Starting up telegram bot...");
    let bot = Bot::new(TG_KEY);
    let me = bot.get_me().await.unwrap();
    log::info!("Logged in to telegram as \"{}\"", me.first_name);
    
    let chat = bot.get_chat(Recipient::Id(ChatId(CHAT_ID.parse::<i64>().unwrap()))).await.unwrap();
    log::info!("Using chat {}", chat.title().unwrap());

    (bot, chat)
}

async fn start_http(state: SharedState) -> Server<SharedState> {
    let mut app = tide::with_state(state);

    app.at("/pull_wh").post(webhook);

    app
}

async fn webhook(mut req: Request<SharedState>) -> tide::Result {

    let body = req.body_string().await.unwrap();
    let serialized = serde_json::from_str::<gitea::PullWh>(body.as_str());

    if serialized.is_err() {
        return Ok(format!("Bad serialization: {}", serialized.unwrap_err().to_string()).into());
    }

    let secret = req.header("Authorization");
    if secret.is_some() {
        // let secret = secret.unwrap().to_string();
        // if ! secret.ends_with(WH_SECRET)
        // return Ok(secret.unwrap().as_str().into());
    }

    Ok(req.body_string().await.unwrap().into())
    // Ok("()".into())
}

#[tokio::main]
async fn main() {
    
    #[cfg(debug_assertions)]
    femme::with_level(log::LevelFilter::Debug);

    #[cfg(not(debug_assertions))] {
        femme::with_level(log::LevelFilter::Info);
        log::info!("Running in production");
    }

    let (bot, chat) = start_tg().await;
    let http = start_http(SharedState { bot, chat }).await;

    http.listen(WH_LISTEN_URL.clone()).await.unwrap();

    let time = POLL_TIME.parse::<u64>().unwrap();

    log::info!("Starting listening for updates ({})", time);
    
    #[allow(dead_code, unreachable_code)]
    loop {

        break;

        let (new_prs, new_data): (Vec<gitea::Pull>, Vec<gitea::Pull>) = (vec![], vec![]);
        log::debug!("Found {} new PRs", new_prs.len());

        if new_prs.len() != 0 {
            let prs = new_data;
            let data = serde_json::to_string_pretty(&prs).unwrap();
            fs::write(".pr-cache", data).await.unwrap();

            for pr in new_prs.iter() {
                log::info!("Sending message about {} ({})", pr.title, pr.url);
                bot.send_message(chat.id, format!("New PR\n{} ({}#{}) by {}\n{}", pr.title, pr.head.repo.name, pr.number, pr.user.login, pr.url)).await.unwrap();
                bot.send_message(chat.id, "апрувните пж @bleki42 @balistiktw @x3paerz").await.unwrap();
            }
        }
    }

}
